package eu.mhsl.minecraft.pixelblocks.pixelblock;

import co.aikar.taskchain.TaskChain;
import eu.mhsl.minecraft.pixelblocks.Main;
import eu.mhsl.minecraft.pixelblocks.PixelBlockItem;
import eu.mhsl.minecraft.pixelblocks.utils.Direction;
import net.kyori.adventure.text.Component;
import net.kyori.adventure.text.format.NamedTextColor;
import org.bukkit.Location;
import org.bukkit.Sound;
import org.bukkit.World;
import org.bukkit.entity.Item;
import org.bukkit.entity.Player;
import org.bukkit.util.Vector;
import org.jetbrains.annotations.NotNull;

import javax.annotation.Nullable;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.UUID;

public class PixelBlock {
    private boolean exists = true;
    private PixelBlockWorld pixelWorld;

    private final Location pixelBlockLocation;
    private final Direction facingDirection;
    private Pixels pixels;
    private List<PixelBlockWorld.PixelData> pixelData;

    private PixelBlockHitbox hitbox;
    private PixelBlockPlaceholder placeholder;

    private Location lastEntryLocation;
    private final UUID ownerUUID;
    private final UUID blockUUID;

    public static @Nullable PixelBlock getPixelBlockFromBlockWorld(World world) {
        return Main.pixelBlocks.stream()
            .filter(block -> block.blockUUID.equals(getUUIDFromWorld(world)))
            .findFirst()
            .orElse(null);
    }

    public static @Nullable UUID getUUIDFromWorld(@NotNull World world) {
        try {
            return UUID.fromString(List.of(world.getName().split("/")).getLast());
        } catch(IllegalArgumentException e) {
            return null;
        }
    }

    public static @Nullable PixelBlock getPixelBlockFromPlacedLocation(@NotNull Location placedLocation) {
        Location searchLocation = placedLocation.clone().toBlockLocation();
        searchLocation.setPitch(0);
        searchLocation.setYaw(0);

        return Main.pixelBlocks.stream()
            .filter(block -> Objects.equals(block.pixelBlockLocation, searchLocation))
            .findFirst()
            .orElse(null);
    }

    public static PixelBlock fromExisting(UUID blockUUID, UUID ownerUUID, Location pixelBlockLocation, Direction direction, Location lastEntryLocation) {
        return new PixelBlock(blockUUID, ownerUUID, pixelBlockLocation, direction, lastEntryLocation);
    }

    private PixelBlock(UUID blockUUID, UUID ownerUUID, Location pixelBlockLocation, Direction direction, Location lastEntryLocation) {
        this.blockUUID = blockUUID;
        this.ownerUUID = ownerUUID;
        this.pixelBlockLocation = pixelBlockLocation;
        this.pixelBlockLocation.setYaw(0);
        this.pixelBlockLocation.setPitch(0);
        this.facingDirection = direction;
        this.lastEntryLocation = lastEntryLocation;

        try {
            this.pixelWorld = new PixelBlockWorld(this);
            this.pixelData = this.pixelWorld.getPixels(this.facingDirection);
            this.pixels = new Pixels(this);
            this.placeholder = new PixelBlockPlaceholder(this);
            this.hitbox = new PixelBlockHitbox(this);

            Main.logger().info(String.format("Loaded existing pixelblock '%s'", this.blockUUID));
        } catch(Exception e) {
            Main.logger().info(String.format("Failed initializing existing pixelblock '%s': %s", this.blockUUID, e.getMessage()));
        }
    }

    public static PixelBlock createPixelBlock(UUID blockUUID, UUID ownerUUID, Location pixelBlockLocation, Direction direction) {
        return new PixelBlock(blockUUID, ownerUUID, pixelBlockLocation, direction);
    }

    private PixelBlock(UUID blockUUID, UUID ownerUUID, Location pixelBlockLocation, Direction direction) {
        if(Main.pixelBlocks.stream().anyMatch(pixelBlock -> pixelBlock.getBlockUUID().equals(blockUUID)))
            throw new IllegalStateException(String.format("PixelBlock '%s' ist bereits in der Welt vorhanden!", blockUUID));

        this.blockUUID = blockUUID;
        this.ownerUUID = ownerUUID;
        this.pixelBlockLocation = pixelBlockLocation;
        this.pixelBlockLocation.setYaw(0);
        this.pixelBlockLocation.setPitch(0);
        this.facingDirection = direction;

        this.getBlockTaskChain()
            .sync(() -> {
                this.pixelWorld = new PixelBlockWorld(this);
                this.pixelData = this.pixelWorld.getPixels(this.facingDirection);
                this.pixels = new Pixels(this);
                this.placeholder = new PixelBlockPlaceholder(this);
                this.hitbox = new PixelBlockHitbox(this);
            })
            .execute();

        this.scheduleEntityUpdate();

        this.getBlockTaskChain()
            .async(() -> {
                Main.database().savePixelBlock(this);
                Main.pixelBlocks.add(this);
            })
            .execute();
    }

    public <T> TaskChain<T> getBlockTaskChain() {
        return Main.sharedChain(this.blockUUID.toString());
    }

    public void enterBlock(@NotNull Player player) {
        if(Main.configuration().onlyEditableByOwner() && !player.getUniqueId().equals(ownerUUID)) {
            player.sendMessage(Component.text("Dieser Pixelblock gehört nicht dir!", NamedTextColor.RED));
            return;
        }

        getBlockTaskChain()
            .async(() -> {
                this.lastEntryLocation = player.getLocation();
                Main.database().savePixelBlock(this);
            })
            .sync(() -> {
                if(!exists) return;
                player.teleport(this.pixelWorld.getSpawnLocation());
            })
            .current(() -> Main.logger()
                .info(String.format("'%s' entered PixelBlock '%s' at %s", player.getName(), this.blockUUID, this.pixelBlockLocation.toString())))
            .execute();
    }

    public void exitBlock(@NotNull Player player) {
        this.getBlockTaskChain()
            .sync(() -> player.teleport(this.lastEntryLocation != null ? this.lastEntryLocation : this.pixelBlockLocation))
            .current(() -> Main.logger().info(String.format("%s exited PixelBlock", player.getName())))
            .sync(() -> this.pixelData = this.pixelWorld.getPixels(this.facingDirection))
            .execute();

        this.scheduleEntityUpdate();
    }

    private void scheduleEntityUpdate() {
        this.getBlockTaskChain()
            .sync(this::removeEntities)
            .async(() -> Collections.shuffle(this.pixelData))
            .delay(1)
            .syncLast((pixelData) -> {
                this.pixels.spawn();
                this.hitbox.spawn();
                this.placeholder.spawn();
            })
            .execute();
    }

//    public void place(Location placeLocation, Direction direction) {
//        Location newLocation = placeLocation.toBlockLocation();
//        newLocation.setPitch(0);
//        newLocation.setYaw(0);
//
//        @Nullable PixelBlock blockAtLocation = PixelBlock.getPixelBlockFromPlacedLocation(newLocation);
//        if(blockAtLocation != null && blockAtLocation != this) throw new IllegalArgumentException("Es können nicht mehrere Pixelblöcke ineinander platziert werden.");
//
//        Main.logger().info(String.format("Placing PixelBlock '%s' at %s", this.blockUUID, placeLocation));
//        this.pixelBlockLocation = newLocation;
//        this.facingDirection = direction;
//        updateEntities();
//        Main.database.savePixelBlock(this);
//
//        this.placeholder = PixelBlockPlaceholder.newPlaceholder(this);
//    }

    public void destroy(Player destroyedBy) {
        if(!this.exists) return;
        if(Main.configuration().onlyBreakableByOwner() && !destroyedBy.getUniqueId().equals(ownerUUID)) {
            destroyedBy.sendMessage("Dieser Pixelblock gehört nicht dir!");
            return;
        }

        this.pixelWorld.getPlayersInWorld().forEach(p -> {
            p.sendMessage(Component.text("Der Pixelblock wurde von einem anderen Spieler abgebaut!", NamedTextColor.RED));
            p.teleport(this.lastEntryLocation);
        });

        Main.logger().info(String.format("Destroying PixelBlock '%s' at %s", this.blockUUID, pixelBlockLocation));
        this.exists = false;

        this.pixelWorld.getEntitiesInWorld().stream()
            .filter(entity -> entity instanceof Item)
            .forEach(entity -> entity.teleport(this.lastEntryLocation));

        this.getBlockTaskChain()
            .sync(() -> {
                this.removeEntities();
                World world = this.pixelBlockLocation.getWorld();
                world.playSound(this.pixelBlockLocation, Sound.BLOCK_COPPER_BULB_BREAK, 1.0F, 30);
                world.dropItem(this.pixelBlockLocation.add(new Vector(0.5, 0.5, 0.5)), PixelBlockItem.getBlockAsItem(this));
            })
            .async(() -> {
                Main.database().deletePixelBlock(this);
                Main.pixelBlocks.remove(this);
            })
            .execute();
    }

    private void removeEntities() {
        this.pixels.destroy();
        this.placeholder.destroy();
        this.hitbox.destroy();
    }

    public @NotNull PixelBlockWorld getPixelWorld() {
        return pixelWorld;
    }

    public Location getPixelBlockLocation() {
        return pixelBlockLocation.clone();
    }

    public Direction getFacingDirection() {
        return facingDirection;
    }

    public boolean hasLastEntryLocation() {
        return this.lastEntryLocation != null;
    }

    public List<PixelBlockWorld.PixelData> getPixelData() {
        return pixelData;
    }

    public Location getLastEntryLocation() {
        return lastEntryLocation.clone();
    }

    public UUID getOwnerUUID() {
        return ownerUUID;
    }

    public UUID getBlockUUID() {
        return blockUUID;
    }
}
